/**
 * Head of code common to all C targets
 **/

#include "beremiz.h"
#include <string.h>
/*
 * Prototypes of functions provided by generated C softPLC
 **/
void config_run__(unsigned long tick);
void config_init__(void);

/*
 * Prototypes of functions provided by generated target C code
 * */
long long AtomicCompareExchange64(long long*, long long, long long);
void __init_debug(void);
void __cleanup_debug(void);
/*void __retrieve_debug(void);*/
void __publish_debug(void);

/*
 *  Variables used by generated C softPLC and plugins
 **/
IEC_TIME __CURRENT_TIME;
IEC_BOOL __DEBUG = 0;
unsigned long __tick = 0;
char *PLC_ID = NULL;

/*
 *  Variable generated by C softPLC and plugins
 **/
extern unsigned long greatest_tick_count__;

/* Help to quit cleanly when init fail at a certain level */
static int init_level = 0;

/*
 * Prototypes of functions exported by plugins
 **/


/*
 * Retrieve input variables, run PLC and publish output variables
 **/
void __run(void)
{
    __tick++;
    if (greatest_tick_count__)
        __tick %= greatest_tick_count__;

    

    /*__retrieve_debug();*/

    config_run__(__tick);

    __publish_debug();

    

}

/*
 * Initialize variables according to PLC's default values,
 * and then init plugins with that values
 **/
int __init(int argc,char **argv)
{
    int res = 0;
    init_level = 0;
    
    /* Effective tick time with 1ms default value */
    if(!common_ticktime__)
        common_ticktime__ = 1000000;

    config_init__();
    __init_debug();
    
    return res;
}
/*
 * Calls plugin cleanup proc.
 **/
void __cleanup(void)
{
    
    __cleanup_debug();
}

void PLC_GetTime(IEC_TIME *CURRENT_TIME);
void PLC_SetTimer(unsigned long long next, unsigned long long period);



/* @page plc_plm2004_main.c
 *       PLM2004 (YAPLC)
 *       platform-dependent code
 *       2020, lamsystems-it.ru
 *       v. 1.0.0
 */

/**
 * Newlib stubs.
 **/

#include <sys/stat.h>

int _close(int file)
{
    (void)file;
    return -1;
}

int _fstat(int file, struct stat *st)
{
    (void)file;
    st->st_mode = S_IFCHR;
    return 0;
}

int _isatty(int file)
{
    (void)file;
    return 1;
}

int _lseek(int file, int ptr, int dir)
{
    (void)file;
    (void)ptr;
    (void)dir;
    return 0;
}

int _open(const char *name, int flags, int mode)
{
    (void)name;
    (void)flags;
    (void)mode;
    return -1;
}

int _read(int file, char *ptr, int len)
{
    (void)file;
    (void)ptr;
    (void)len;
    return 0;
}

char *heap_end = 0;
caddr_t _sbrk(int incr)
{
    (void)incr;
    return (caddr_t) 0;
}

int _write(int file, char *ptr, int len)
{
    (void)file;
    (void)ptr;
    (void)len;
    return 0;
}

/**
 * YAPLC specific code.
 **/

#include <string.h>
#include <iec_std_lib.h>
#include <plc_abi.h>

/**
 * YAPLC ABI.
 **/

void fake_start(void)
{
    while(1);
}

/*!< TODO: добавть специальную секцию для
type * name = &(PLC_LOC_BUF(name));
Чтобы можно было разместть их во flash-памяти.
*/

/* LOCATED_VARIABLES
 * global variables
 */

#define PLC_LOC_BUF(name)  PLC_LOC_CONCAT(name, _BUF)
#define PLC_LOC_ADDR(name) PLC_LOC_CONCAT(name, _ADDR)
#define PLC_LOC_DSC(name)  PLC_LOC_CONCAT(name, _LDSC)

#define __LOCATED_VAR( type, name, lt, lsz, io_proto, ... ) \
type PLC_LOC_BUF(name);                                     \
type * name = &(PLC_LOC_BUF(name));                         \
const uint32_t PLC_LOC_ADDR(name)[] = {__VA_ARGS__};        \
const plc_loc_dsc_t PLC_LOC_DSC(name) =                     \
    {                                                       \
     .v_buf  = (void *)&(PLC_LOC_BUF(name)),                \
     .v_type = PLC_LOC_TYPE(lt),                            \
     .v_size = PLC_LOC_SIZE(lsz),                           \
     .a_size = sizeof(PLC_LOC_ADDR(name))/sizeof(uint32_t), \
     .a_data = &(PLC_LOC_ADDR(name)[0]),                    \
     .proto  = io_proto                                     \
    };

#include "LOCATED_VARIABLES.h"
#include "LOCATED_VARIABLES_HIDDEN.h"
#undef __LOCATED_VAR


/** LOCATED VARIABLES
 */

#define __LOCATED_VAR(type, name, ...) &(PLC_LOC_DSC(name)),
plc_loc_tbl_t plc_loc_table[] =
{
#include "LOCATED_VARIABLES.h"
#include "LOCATED_VARIABLES_HIDDEN.h"
};
#undef __LOCATED_VAR

#define PLC_LOC_TBL_SIZE (sizeof(plc_loc_table)/sizeof(plc_loc_dsc_t *))
uint32_t plc_loc_weigth[PLC_LOC_TBL_SIZE];



/** LOCATED VARIABLES
 *  EEPROM-registers
 */

#define __LOCATED_VAR(AddrIn, TyIn, ... ) \
    {                                     \
     .Addr = AddrIn,                      \
     .Ty   = TyIn,                        \
     .Rex  = FALSE,                       \
     .Rval = 0.0,                         \
     .Rok  = FALSE,                       \
     .Wex  = FALSE,                       \
     .Wval = 0.0,                         \
     .Wok  = FALSE                        \
    },
plc_app_eereg_t PlcEe_Regs[] =
{
#include "LOCATED_VARIABLES_EEREG.h"
};
#undef __LOCATED_VAR

#define PLC_EE_REGS_SIZE (sizeof(PlcEe_Regs)/sizeof(plc_app_eereg_t))



/** LOCATED FUNCTIONS
 */

plc_app_funcs_t PlcAppFuncs = 
{
    .DIMode     = 0,
    .DIVal      = 0,
    .DICntT     = 0,
    .DICnt      = 0,
    .DICntRst   = 0,
    .DIEnc      = 0,
    
    .AIMode     = 0,
    .AIVal      = 0,
    
    .DTMode     = 0,
    .DTSetID    = 0,
    .DTGetID    = 0,
    .DTVal      = 0,
    
    .PTMode     = 0,
    .PTVal      = 0,
    
    .DOMode     = 0,
    .DOVal      = 0,
    .DOFast     = 0,
    .DOPwm      = 0,
    .DOSafe     = 0,
    
    .AOMode     = 0,
    .AOVal      = 0,
    .AOFast     = 0,
    .AOSafe     = 0,
    
    .SoftReset  = 0,
    
    .GetTime    = 0
};



#ifndef PLC_MD5
#error "PLC_MD5 must be defined!!!"
#endif

#define PLC_MD5_STR(a) # a
#define PLC_MD5_STR2(a) PLC_MD5_STR(a)
//App ABI, placed after .plc_app_abi_sec
__attribute__ ((section(".plc_md5_sec"))) char plc_md5[] = PLC_MD5_STR2(PLC_MD5);
//App ABI, placed at the .text end
__attribute__ ((section(".plc_check_sec"))) char plc_check_md5[] = PLC_MD5_STR2(PLC_MD5);

//Linker added symbols
extern uint32_t _plc_data_loadaddr, _plc_data_start, _plc_data_end, _plc_bss_end, _plc_sstart;

extern app_fp_t _plc_pa_start, _plc_pa_end;
extern app_fp_t _plc_ia_start, _plc_ia_end;
extern app_fp_t _plc_fia_start,_plc_fia_end;

extern int startPLC(int argc,char **argv);
extern int stopPLC();
extern void runPLC(void);

extern void resumeDebug(void);
extern void suspendDebug(int disable);

extern void FreeDebugData(void);
extern int GetDebugData(unsigned long *tick, unsigned long *size, void **buffer);

extern void ResetDebugVariables(void);
extern void RegisterDebugVariable(int idx, void* force);

extern void ResetLogCount(void);
extern uint32_t GetLogCount(uint8_t level);
extern uint32_t GetLogMessage(uint8_t level, uint32_t msgidx, char* buf, uint32_t max_size, uint32_t* tick, uint32_t* tv_sec, uint32_t* tv_nsec);

//App ABI, placed at the .text start
__attribute__ ((section(".plc_app_abi_sec"))) plc_app_abi_t plc_yaplc_app =
{
    .sstart = (uint32_t *)&_plc_sstart,
    .entry  = fake_start,
    
    //Startup interface
    .data_loadaddr = &_plc_data_loadaddr,

    .data_start    = &_plc_data_start,
    .data_end      = &_plc_data_end,

    .bss_end       = &_plc_bss_end,

    .pa_start  = &_plc_pa_start,
    .pa_end    = &_plc_pa_end,

    .ia_start  = &_plc_ia_start,
    .ia_end    = &_plc_ia_end,

    .fia_start = &_plc_fia_start,
    .fia_end   = &_plc_fia_end,

    .check_id  = plc_check_md5,

    //Must be run on compatible RTE
    .rte_ver_major = 1,
    .rte_ver_minor = 0,
    .rte_ver_patch = 0,
    
    .hw_id = 2004,
    
    //Located Variables
    // IO manager interface
    .l_tab = &plc_loc_table[0],
    .w_tab = &plc_loc_weigth[0],
    .l_sz  = PLC_LOC_TBL_SIZE,
    
    //EEPROM-registers
    .ee_regs = &PlcEe_Regs[0],
    .ee_sz   = PLC_EE_REGS_SIZE,
    
    //Located functions
    .funcs = &PlcAppFuncs,

    //Common ticktime
    .common_ticktime = &common_ticktime__,

    //App interface
    .id   = plc_md5,

    .start = startPLC,
    .stop  = stopPLC,
    .run   = runPLC,

    .dbg_resume    = resumeDebug,
    .dbg_suspend   = suspendDebug,

    .dbg_data_get  = GetDebugData,
    .dbg_data_free = FreeDebugData,

    .dbg_vars_reset   = ResetDebugVariables,
    .dbg_var_register = RegisterDebugVariable,

    .log_cnt_get   = GetLogCount,
    .log_msg_get   = GetLogMessage,
    .log_cnt_reset = ResetLogCount,
    .log_msg_post  = LogMessage
};

//Redefine LOG_BUFFER_SIZE
#define LOG_BUFFER_SIZE (1<<10) /*1Ko*/
#define LOG_BUFFER_ATTRS

#define PLC_RTE ((plc_rte_abi_t *)(PLC_RTE_ADDR))

void PLC_GetTime(IEC_TIME *CURRENT_TIME)
{
    //PLC_RTE->get_time( CURRENT_TIME );
    if(PlcAppFuncs.GetTime)
    {
       PlcAppFuncs.GetTime(CURRENT_TIME);
    }
    (void)CURRENT_TIME;
}

void PLC_SetTimer(unsigned long long next, unsigned long long period)
{
    //@TODO ???
    //PLC_RTE->set_timer( next, period );
    (void)next;
    (void)period;
}

long AtomicCompareExchange(long* atomicvar,long compared, long exchange)
{
	/* No need for real atomic op on LPC,
	 * no possible preemption between debug and PLC */
	long res = *atomicvar;
	if(res == compared){
		*atomicvar = exchange;
	}
	return res;
}

long long AtomicCompareExchange64(long long* atomicvar,long long compared, long long exchange)
{
	/* No need for real atomic op on LPC,
	 * no possible preemption between debug and PLC */
	long long res = *atomicvar;
	if(res == compared){
		*atomicvar = exchange;
	}
	return res;
}

static int debug_locked = 0;
static int _DebugDataAvailable = 0;
static unsigned long __debug_tick;

int TryEnterDebugSection(void)
{
    if(!debug_locked && __DEBUG){
        debug_locked = 1;
		return 1;
    }
    return 0;
}

void LeaveDebugSection(void)
{
        debug_locked = 0;
}

void InitiateDebugTransfer(void)
{
    /* remember tick */
    __debug_tick = __tick;
    _DebugDataAvailable = 1;
}

void suspendDebug(int disable)
{
    /* Prevent PLC to enter debug code */
    __DEBUG = !disable;
    debug_locked = !disable;
}

void resumeDebug(void)
{
    /* Let PLC enter debug code */
    __DEBUG = 1;
    debug_locked = 0;
}

int WaitDebugData(unsigned long *tick)
{
    if(_DebugDataAvailable && !debug_locked){
        /* returns 0 on success */
        *tick = __debug_tick;
        _DebugDataAvailable = 0;
        return 0;
    }
    return 1;
}

void ValidateRetainBuffer(void)
{
    //PLC_RTE->validate_retain_buf();
}
void InValidateRetainBuffer(void)
{
    //PLC_RTE->invalidate_retain_buf();
}
int CheckRetainBuffer(void)
{
    //return PLC_RTE->check_retain_buf();
    return (1);
}

void InitRetain(void)
{
}

void CleanupRetain(void)
{
}

void Retain(unsigned int offset, unsigned int count, void *p)
{
    //PLC_RTE->retain( offset, count, p );
}
void Remind(unsigned int offset, unsigned int count, void *p)
{
    //PLC_RTE->remind( offset, count, p );
}

int startPLC(int argc,char **argv)
{
	if(__init(argc,argv) == 0){
		PLC_SetTimer(0, common_ticktime__);
		return 0;
	}else{
		return 1;
	}
}

int stopPLC(void)
{
    __cleanup();
    return 0;
}

void runPLC(void)
{
    PLC_GetTime( &__CURRENT_TIME );
    __run();
}



/** F(B) C-code
 */

#include "iec_types_all.h"
#include "POUS.h"

//@TODO + define APP_* from LOCATED_VARIABLES_HIDDEN (by used libraries)

#define LIB_APP_DI
#define LIB_APP_DO
#define LIB_APP_AO
#define LIB_APP_DT
#define LIB_APP_PT
#define LIB_APP_AI
#define LIB_APP_EEPROM
#define LIB_APP_FLR
#define LIB_APP_MBRTU_MASTER
#define LIB_APP_SYS

/** DI
 */

#ifdef LIB_APP_DI

/** @brief  DIMode.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DIMode(DIMODE *DataIn)
{
    if(DataIn && PlcAppFuncs.DIMode)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE DIn = __GET_VAR(DataIn->DIN);
        BYTE M   = __GET_VAR(DataIn->M);
        BYTE Om  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.DIMode(&Ex, &DIn, &M, &Om, &Ok);
        
        __SET_VAR(DataIn->, OM,, Om);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  DIVal.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DIVal(DIVAL *DataIn)
{
    if(DataIn && PlcAppFuncs.DIVal)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE DIn = __GET_VAR(DataIn->DIN);
        BOOL Ov  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.DIVal(&Ex, &DIn, &Ov, &Ok);
        
        __SET_VAR(DataIn->, OV,, Ov);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  DICntT.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DICntT(DICNTT *DataIn)
{
    if(DataIn && PlcAppFuncs.DICntT)
    {
        BOOL Ex    = __GET_VAR(DataIn->EX);
        BYTE DIn   = __GET_VAR(DataIn->DIN);
        WORD Ref   = __GET_VAR(DataIn->REF);
        BOOL RefEn = __GET_VAR(DataIn->REFEN);
        WORD Ov    = 0;
        BOOL Oref  = 0;
        BYTE Ok    = 0;
        
        PlcAppFuncs.DICntT(&Ex, &DIn, &Ref, &RefEn, &Ov, &Oref, &Ok);
        
        __SET_VAR(DataIn->, OV,, Ov);
        __SET_VAR(DataIn->, OREF,, Oref);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  DICnt.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DICnt(DICNT *DataIn)
{
    if(DataIn && PlcAppFuncs.DICnt)
    {
        BOOL Ex    = __GET_VAR(DataIn->EX);
        BYTE DIn   = __GET_VAR(DataIn->DIN);
        DWORD Ref  = __GET_VAR(DataIn->REF);
        BOOL RefEn = __GET_VAR(DataIn->REFEN);
        DWORD Ov   = 0;
        BOOL Oref  = 0;
        BYTE Ok    = 0;
        
        PlcAppFuncs.DICnt(&Ex, &DIn, &Ref, &RefEn, &Ov, &Oref, &Ok);
        
        __SET_VAR(DataIn->, OV,, Ov);
        __SET_VAR(DataIn->, OREF,, Oref);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  DICntRst.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DICntRst(DICNTRST *DataIn)
{
    if(DataIn && PlcAppFuncs.DICntRst)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE DIn = __GET_VAR(DataIn->DIN);
        BOOL Rst = __GET_VAR(DataIn->RST);
        BYTE Ok  = 0;
        
        PlcAppFuncs.DICntRst(&Ex, &DIn, &Rst, &Ok);
        
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  DIEnc.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DIEnc(DIENC *DataIn)
{
    if(DataIn && PlcAppFuncs.DIEnc)
    {
        BOOL  Ex     = __GET_VAR(DataIn->EX);
        BYTE  DIn    = __GET_VAR(DataIn->DIN);
        DWORD RefA   = __GET_VAR(DataIn->REFA);
        BOOL  RefAEn = __GET_VAR(DataIn->REFAEN);
        DWORD RefB   = __GET_VAR(DataIn->REFB);
        BOOL  RefBEn = __GET_VAR(DataIn->REFBEN);
        WORD  RefT   = __GET_VAR(DataIn->REFT);
        BOOL  RefTEn = __GET_VAR(DataIn->REFTEN);
        DWORD OA     = 0;  
        BOOL  OAref  = 0;
        DWORD OB     = 0;
        BOOL  OBref  = 0;
        WORD  OT     = 0;
        BOOL  OTref  = 0;
        BYTE  Ok     = 0;
        
        PlcAppFuncs.DIEnc(&Ex, &DIn, &RefA, &RefAEn, &RefB, &RefBEn, &RefT, &RefTEn, &OA, &OAref, &OB, &OBref, &OT, &OTref, &Ok);
        
        __SET_VAR(DataIn->, OA,, OA);
        __SET_VAR(DataIn->, OAREF,, OAref);
        __SET_VAR(DataIn->, OB,, OB);
        __SET_VAR(DataIn->, OBREF,, OBref);
        __SET_VAR(DataIn->, OT,, OT);
        __SET_VAR(DataIn->, OTREF,, OTref);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

#endif // LIB_APP_DI



/** AI
 */

#ifdef LIB_APP_AI

/** @brief  AIMode.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_AIMode(AIMODE *DataIn)
{
    if(DataIn && PlcAppFuncs.AIMode)
    {
        BOOL  Ex  = __GET_VAR(DataIn->EX);
        BYTE  AIn = __GET_VAR(DataIn->AIN);
        BYTE  M   = __GET_VAR(DataIn->M);
        BYTE Om   = 0;
        BYTE Os   = 0;
        BYTE Ok   = 0;
        
        PlcAppFuncs.AIMode(&Ex, &AIn, &M, &Om, &Os, &Ok);
        
        __SET_VAR(DataIn->, OM,, Om);
        __SET_VAR(DataIn->, OS,, Os);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  AIVal.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_AIVal(AIVAL *DataIn)
{
    if(DataIn && PlcAppFuncs.AIVal)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE AIn = __GET_VAR(DataIn->AIN);
        REAL Ov  = 0.0;
        BYTE Os  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.AIVal(&Ex, &AIn, &Ov, &Os, &Ok);
        
        __SET_VAR(DataIn->, OV,, Ov);
        __SET_VAR(DataIn->, OS,, Os);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

#endif // LIB_APP_AI



/** DT
 */

#ifdef LIB_APP_DT

/** @brief  DTMode.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DTMode(DTMODE *DataIn)
{
    if(DataIn && PlcAppFuncs.DTMode)
    {
        BOOL  Ex  = __GET_VAR(DataIn->EX);
        BYTE  DTn = __GET_VAR(DataIn->DTN);
        BYTE  M   = __GET_VAR(DataIn->M);
        BYTE Om   = 0;
        BYTE Os   = 0;
        BYTE Ok   = 0;
        
        PlcAppFuncs.DTMode(&Ex, &DTn, &M, &Om, &Os, &Ok);
        
        __SET_VAR(DataIn->, OM,, Om);
        __SET_VAR(DataIn->, OS,, Os);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  DTSetId.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DTSetId(DTSETID *DataIn)
{
    if(DataIn && PlcAppFuncs.DTSetID)
    {
        BOOL  Ex  = __GET_VAR(DataIn->EX);
        BYTE  DTn = __GET_VAR(DataIn->DTN);
        LWORD ID  = __GET_VAR(DataIn->ID);
        LWORD OID = 0;
        BYTE  Ok  = 0; 
        
        PlcAppFuncs.DTSetID(&Ex, &DTn, &ID, &OID, &Ok);
        
        __SET_VAR(DataIn->, OID,, OID);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  DTGetId.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DTGetId(DTGETID *DataIn)
{
    if(DataIn && PlcAppFuncs.DTGetID)
    {
        BOOL  Ex  = __GET_VAR(DataIn->EX);
        BYTE  DTn = __GET_VAR(DataIn->DTN);
        LWORD OID = 0;
        BYTE  Ok  = 0; 
        
        PlcAppFuncs.DTGetID(&Ex, &DTn, &OID, &Ok);
        
        __SET_VAR(DataIn->, OID,, OID);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  DTVal.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DTVal(DTVAL *DataIn)
{
    if(DataIn && PlcAppFuncs.DTVal)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE DTn = __GET_VAR(DataIn->DTN);
        REAL Ov  = 0.0;
        BYTE Os  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.DTVal(&Ex, &DTn, &Ov, &Os, &Ok);
        
        __SET_VAR(DataIn->, OV,, Ov);
        __SET_VAR(DataIn->, OS,, Os);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

#endif // LIB_APP_DT



/** PT
 */

#ifdef LIB_APP_PT

/** @brief  PTMode.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_PTMode(PTMODE *DataIn)
{
    if(DataIn && PlcAppFuncs.PTMode)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE PTn = __GET_VAR(DataIn->PTN);
        BYTE M   = __GET_VAR(DataIn->M);
        BYTE STy = __GET_VAR(DataIn->STY);
        BYTE Om  = 0;
        BYTE Os  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.PTMode(&Ex, &PTn, &M, &STy, &Om, &Os, &Ok);
        
        __SET_VAR(DataIn->, OM,, Om);
        __SET_VAR(DataIn->, OS,, Os);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  PTVal.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_PTVal(PTVAL *DataIn)
{
    if(DataIn && PlcAppFuncs.PTVal)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE PTn = __GET_VAR(DataIn->PTN);
        REAL Ov  = 0.0;
        BYTE Os  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.PTVal(&Ex, &PTn, &Ov, &Os, &Ok);
        
        __SET_VAR(DataIn->, OV,, Ov);
        __SET_VAR(DataIn->, OS,, Os);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

#endif // LIB_APP_PT



/** DO
 */

#ifdef LIB_APP_DO

/** @brief  DOMode.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DOMode(DOMODE *DataIn)
{
    if(DataIn && PlcAppFuncs.DOMode)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE DOn = __GET_VAR(DataIn->DON);
        BYTE M   = __GET_VAR(DataIn->M);
        BYTE Om  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.DOMode(&Ex, &DOn, &M, &Om, &Ok);
        
        __SET_VAR(DataIn->, OM,, Om);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  DOVal.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DOVal(DOVAL *DataIn)
{
    if(DataIn && PlcAppFuncs.DOVal)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE DOn = __GET_VAR(DataIn->DON);
        BOOL V   = __GET_VAR(DataIn->V);
        BOOL Ov  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.DOVal(&Ex, &DOn, &V, &Ov, &Ok);
        
        __SET_VAR(DataIn->, OV,, Ov);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  DOFast.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DOFast(DOFAST *DataIn)
{
    if(DataIn && PlcAppFuncs.DOFast)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE DOn = __GET_VAR(DataIn->DON);
        BOOL V   = __GET_VAR(DataIn->V);
        BOOL Ov  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.DOFast(&Ex, &DOn, &V, &Ov, &Ok);
        
        __SET_VAR(DataIn->, OV,, Ov);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  DOPwm.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DOPwm(DOPWM *DataIn)
{
    if(DataIn && PlcAppFuncs.DOPwm)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BOOL En  = __GET_VAR(DataIn->ENX);
        BYTE DOn = __GET_VAR(DataIn->DON);
        DWORD Tm = __GET_VAR(DataIn->TM);
        REAL  D  = __GET_VAR(DataIn->DU);
        BOOL Ow  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.DOPwm(&Ex, &En, &DOn, &Tm, &D, &Ow, &Ok);
        
        __SET_VAR(DataIn->, OW,, Ow);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  DOSafe.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_DOSafe(DOSAFE *DataIn)
{
    if(DataIn && PlcAppFuncs.DOSafe)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BOOL En  = __GET_VAR(DataIn->EN);
        BYTE DOn = __GET_VAR(DataIn->DON);
        BOOL V   = __GET_VAR(DataIn->V);
        BOOL Ov  = 0;
        BOOL Ow  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.DOSafe(&Ex, &En, &DOn, &V, &Ov, &Ow, &Ok);
        
        __SET_VAR(DataIn->, OV,, Ov);
        __SET_VAR(DataIn->, OW,, Ow);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

#endif // LIB_APP_DO



/** AO
 */

#ifdef LIB_APP_AO

/** @brief  AOMode.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_AOMode(AOMODE *DataIn)
{
    if(DataIn && PlcAppFuncs.AOMode)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE AOn = __GET_VAR(DataIn->AON);
        BYTE M   = __GET_VAR(DataIn->M);
        BYTE Om  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.AOMode(&Ex, &AOn, &M, &Om, &Ok);
        
        __SET_VAR(DataIn->, OM,, Om);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  AOVal.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_AOVal(AOVAL *DataIn)
{
    if(DataIn && PlcAppFuncs.AOVal)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE AOn = __GET_VAR(DataIn->AON);
        REAL V   = __GET_VAR(DataIn->V);
        REAL Ov  = 0.0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.AOVal(&Ex, &AOn, &V, &Ov, &Ok);
        
        __SET_VAR(DataIn->, OV,, Ov);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  AOFast.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_AOFast(AOFAST *DataIn)
{
    if(DataIn && PlcAppFuncs.AOFast)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BYTE AOn = __GET_VAR(DataIn->AON);
        REAL V   = __GET_VAR(DataIn->V);
        REAL Ov  = 0.0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.AOFast(&Ex, &AOn, &V, &Ov, &Ok);
        
        __SET_VAR(DataIn->, OV,, Ov);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

/** @brief  AOSafe.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void App_AOSafe(AOSAFE *DataIn)
{
    if(DataIn && PlcAppFuncs.AOSafe)
    {
        BOOL Ex  = __GET_VAR(DataIn->EX);
        BOOL En  = __GET_VAR(DataIn->EN);
        BYTE AOn = __GET_VAR(DataIn->AON);
        REAL V   = __GET_VAR(DataIn->V);
        REAL Ov  = 0.0;
        BOOL Ow  = 0;
        BYTE Ok  = 0;
        
        PlcAppFuncs.AOSafe(&Ex, &En, &AOn, &V, &Ov, &Ow, &Ok);
        
        __SET_VAR(DataIn->, OV,, Ov);
        __SET_VAR(DataIn->, OW,, Ow);
        __SET_VAR(DataIn->, OK,, Ok);
    }
}

#endif // LIB_APP_AO



/** EEPROM
 */

#ifdef LIB_APP_EEPROM

/** @brief  Clear all memory.
 *  @param  ExIn - allow execution.
 *  @param  OkOut - pointer to result.
 *  @return None.
 */
void AppEe_Clr(BOOL ExIn, BOOL *OkOut)
{
    if(OkOut)
    {
        *__QX4_1_1 = ExIn;
        *OkOut = __IX4_1_5_BUF;
    }
}

/** @brief  Search item in global array PlcEe_Regs.
 *  @param  AddrIn - start address of memory,
 *  @param  TyIn - ID of data type.
 *  @return Index to array PlcEe_Regs or -1.
 */
int32_t AppEe_Search(DINT AddrIn, BYTE TyIn)
{
    int32_t i=0;
    for(i=0; i<PLC_EE_REGS_SIZE; i++)
    {
        if(PlcEe_Regs[i].Addr == AddrIn && PlcEe_Regs[i].Ty == TyIn) return (i);
    }
    return ((int32_t)(-1));
}

/** @brief  Read data.
 *  @param  DataIn - pointer to structure of arguments.
 *  @return None.
 */
void AppEe_Read(EEREADLREAL *DataIn)
{
    if(DataIn)
    {
        EEREGLREAL TEMP = __GET_VAR(DataIn->REG);
        int32_t IDx     = AppEe_Search(TEMP.ADDR, TEMP.TY);
        
        if(IDx > -1)
        {
            PlcEe_Regs[IDx].Rex = __GET_VAR(DataIn->EX);
            
            __SET_VAR(DataIn->, OVAL,, PlcEe_Regs[IDx].Rval);
            __SET_VAR(DataIn->, OK,, PlcEe_Regs[IDx].Rok);
        }
    }
}

/** @brief  Write data.
 *  @param  DataIn - pointer to structure of arguments.
 *  @return None.
 */
void AppEe_Write(EEWRITELREAL *DataIn)
{
    if(DataIn)
    {
        EEREGLREAL TEMP = __GET_VAR(DataIn->REG);
        int32_t IDx     = AppEe_Search(TEMP.ADDR, TEMP.TY);
        
        if(IDx > -1)
        {
            PlcEe_Regs[IDx].Wex  = __GET_VAR(DataIn->EX);
            PlcEe_Regs[IDx].Wval = __GET_VAR(DataIn->VAL);
            
            __SET_VAR(DataIn->, OK,, PlcEe_Regs[IDx].Wok);
        }
    }
}

#endif // LIB_APP_EEPROM



/** FLR
 */

#ifdef LIB_APP_FLR

/** @def PI
 */
#define APP_FLR_PI          (REAL)3.14159265358979

/** @def Quantity of terms
 */
#define APP_FLR_SZ          5

/** @def Input terms
 */
#define APP_FLR_IN_HI2      0   //very higher
#define APP_FLR_IN_HI       1   //high
#define APP_FLR_IN_NORM     2   //normal
#define APP_FLR_IN_LO       3   //low
#define APP_FLR_IN_LO2      4   //very lowest

/** @def Output terms
 */
#define APP_FLR_OUT_OP2     0   //fast open  (speed 2)
#define APP_FLR_OUT_OP1     1   //slow open  (speed 1)
#define APP_FLR_OUT_NONE    2   //nothing
#define APP_FLR_OUT_CL1     3   //slow close (speed 1)
#define APP_FLR_OUT_CL2     4   //fast close (speed 2)

/** @def Direction of output value increase
 */
#define APP_FLR_DIR_DOWN    (SINT)-1
#define APP_FLR_DIR_NONE    (SINT)0
#define APP_FLR_DIR_UP      (SINT)1

/** @def Min/Max
 */
#define APP_FLR_MIN         (REAL)-1.0
#define APP_FLR_MAX         (REAL)1.0

/** @def Inverse probability of normal state
 *  @note q = (1 - Actions[APP_FLR_IN_NORM])
 */
#define APP_FLR_Q_MIN       (REAL)0.0
#define APP_FLR_Q_MAX       (REAL)0.3

/** @def Sigma
 */
#define APP_FLR_SIGMA       (REAL)0.0

/** @def Step of accumulating value increase
 */
#define APP_FLR_DX          (REAL)0.01


/** @brief  Gaus-function.
 *  @param  C   - value of In, where Gaus-function is maximum (1.0).
 *  @param  S   - incremental step of Gaus-function.
 *  @param  In  - input value.
 *  @param  Out - result.
 *  @return Result.
 */
REAL AppFlr_GausFx(const REAL C, const REAL S, const REAL In)
{
    if(S) return (exp(((-1.0)*(0.5*powf((In-C), 2.0)))/(powf(S, 2.0))));
    return (0.0);
}

/** @brief  Triangle-function.
 *  @param  A   - A-argument.
 *  @param  B   - B-argument.
 *  @param  C   - C-argument.
 *  @param  In  - input value.
 *  @return Result.
 */
REAL AppFlr_TriangleFx(const REAL A, const REAL B, const REAL C, const REAL In)
{
    if(In > A && In < C) return ((In < B) ? (In-A)/(B-A) : (C-In)/(C-B));
    return (0.0);
}

/** @brief  Get difference between value and setpoint in Term-format (-1 .. 1).
 *  @param  ErSrc  - source difference (error).
 *  @return None.
 */
REAL AppFlr_ErTerm(const REAL ErSrc)
{
    //trigonometric function
    REAL Er = (270.0-ErSrc);
    if(Er < 180.0) Er = 180.0;  //to min
    if(Er > 360.0) Er = 360.0;  //to max
    Er = Er*(APP_FLR_PI/180.0); //to radian
    Er = cos(Er)*2.0;
    return (Er);
}

/** @brief  Get output terms.
 *  @param  Sigma - Sigma-value.
 *  @param  Min - minimum.
 *  @param  Max - maximum.
 *  @param  In  - input value.
 *  @param  Out - pointer to array of output terms.
 *  @return None.
 */
void AppFlr_OutTerms(const REAL Sigma, const REAL Min, const REAL Max, const REAL In, REAL *Out)
{
    if(Out)
    {
        REAL S = ((Sigma != 0.0) ? Sigma : (Max-Min)/((REAL)(2*APP_FLR_SZ)));
        REAL C;
        BYTE i;
        
        for(i=0; i<APP_FLR_SZ; i++)
        {
            if(In < Min)
            {
                //fast open
                Out[i] = ((i == APP_FLR_OUT_OP2) ? APP_FLR_MAX : 0.0);
            }
            else if(In > Max)
            {
                //fast close
                Out[i] = ((i == APP_FLR_OUT_CL2) ? APP_FLR_MAX : 0.0);
            }
            else
            {
                C = Min+((((REAL)i)*(Max-Min))/((REAL)(APP_FLR_SZ-1)));
                Out[i] = AppFlr_GausFx(C, S, In);
            }
        }
    }
}

/** @brief  Accumulate actions.
 *  @param  dX  - step of value increase.
 *  @param  Min - minimum.
 *  @param  Max - maximum.
 *  @param  Actions - pointer to array of actions.
 *  @return Value of control action.
 */
REAL AppFlr_AccActions(const REAL dX, const REAL Min, const REAL Max, REAL *Actions, REAL *OCm, REAL *Om)
{
    if(Actions && dX)
    {
        /** @var Arguments of Triangle-function
         */
        REAL A[APP_FLR_SZ] = {-1.0, -0.7, -0.25, 0.4, 0.8};
        REAL B[APP_FLR_SZ] = {-0.9, -0.5,  0.0,  0.5, 0.9};
        REAL C[APP_FLR_SZ] = {-0.8, -0.3,  0.25, 0.7, 1.0};
        
        REAL Val, Acc;
        REAL L  = 0.0;
        REAL R  = 0.0;
        REAL cm = 0.0;
        REAL m  = 0.000000001;
        BYTE i;
        
        for(Val=Min; Val<=Max; Val+=dX)
        {
            //accumulate value
            for(i=0; i<APP_FLR_SZ; i++)
            {
                Acc = Actions[i]*AppFlr_TriangleFx(A[i], B[i], C[i], Val);
                if(!i || Acc > R) R = Acc;
            }
            
            cm = cm	+ (dX*(((Val*L)+(Val+dX)*R)/2.0));
            m  = m	+ (((L+R)/2.0)*dX);
            L  = R;
        }
        
        if(OCm) *OCm = cm;
        if(Om)  *Om  = m;
        
        if(m != 0.0) return (cm/m);
    }
    
    return (0.0);
}

/** @brief  Probability of normal state.
 *  @param  V       - current value.
 *  @param  Sp      - setpont.
 *  @param  Speed   - speed of changing V at time TmCy.
 *  @param  ActNorm - value of normal action.
 *  @param  Y       - Y-value.
 *  @return None.
 */
void AppFlr_ProbNorm(const REAL V, const REAL Sp, const REAL Speed, const REAL ActNorm, REAL *Y)
{
    if(Y)
    {
        /** @var Inverse probability of normal state
         */
        REAL q = (1.0-ActNorm);
        if(q > APP_FLR_Q_MIN && q < APP_FLR_Q_MAX)
        {
            if((V < Sp && Speed > 0.0) || (V > Sp && Speed < 0.0))
            {
                *Y *= (-1.0);
            }
        }
    }
}

/** @brief  Get direction of changing Y.
 *  @param  Y - Y-value.
 *  @return UpDown-value.
 */
SINT AppFlr_UpDown(const REAL Y)
{
    if(Y < 0.0) return (APP_FLR_DIR_DOWN);
    if(Y > 0.0) return (APP_FLR_DIR_UP);
    return (APP_FLR_DIR_NONE);
}

/** @brief  Get time of changing Y.
 *  @param  Tm - time of main-cycle (msec).
 *  @param  Y - Y-value.
 *  @return UpDown-value.
 */
REAL AppFlr_dVTm(const REAL Tm, const REAL Y)
{
    if(Tm)
    {
        REAL Yabs = (REAL)fabs((LREAL)Y);
        return ((Tm-(Yabs*Tm)));
    }
    return (0.0);
}

/** @brief  FuzzyLogic.
 *  @param  DataIn - FB-arguments.
 *  @return None.
 */
void AppFlr(FLR1 *DataIn)
{
    if(DataIn)
    {
        //get FB-data
        REAL V      = __GET_VAR(DataIn->V);
        REAL Vprev  = __GET_VAR(DataIn->VPREV);
        REAL Sp     = __GET_VAR(DataIn->SP);
        REAL ErSrc  = __GET_VAR(DataIn->ER);
        REAL Tm     = __GET_VAR(DataIn->TM);
        
        /** @var Debug
         */
        REAL Ocm, Om;
        
        /** @var Speed of changing source value
         */
        REAL Speed = (V-Vprev);
        
        /** @var Difference between value and setpoint (Term-format)
         */
        REAL Er = AppFlr_ErTerm(ErSrc);
        
        /** @var Output terms
         */
        REAL Outputs[APP_FLR_SZ];
        //init. output terms
        AppFlr_OutTerms(APP_FLR_SIGMA, APP_FLR_MIN, APP_FLR_MAX, Er, Outputs);
        
        /** @var Actions
         */
        REAL Actions[APP_FLR_SZ];
        //rules
        Actions[APP_FLR_IN_HI2]  = Outputs[APP_FLR_OUT_CL2];
        Actions[APP_FLR_IN_HI]   = Outputs[APP_FLR_OUT_CL1];
        Actions[APP_FLR_IN_NORM] = Outputs[APP_FLR_OUT_NONE];
        Actions[APP_FLR_IN_LO]   = Outputs[APP_FLR_OUT_OP1];
        Actions[APP_FLR_IN_LO2]  = Outputs[APP_FLR_OUT_OP2];
        //accumulate actions
        REAL Y = AppFlr_AccActions(APP_FLR_DX, APP_FLR_MIN, APP_FLR_MAX, Actions, &Ocm, &Om);
        //probability of normal action
        AppFlr_ProbNorm(V, Sp, Speed, Actions[APP_FLR_IN_NORM], &Y);
        
        /** @var Direction of changing Y
         */
        SINT UpDown = AppFlr_UpDown(Y);
        
        /** @var Time of changing Y (msec)
         */
        REAL TmDv = AppFlr_dVTm(Tm, Y);
        
        //set FB-data
        __SET_VAR(DataIn->, VPREV,, V);
        __SET_VAR(DataIn->, TMDV,, TmDv);
        __SET_VAR(DataIn->, UPDOWN,, UpDown);
        //__SET_VAR(DataIn->, DEB_SPEED,, Speed);
        //__SET_VAR(DataIn->, DEB_Y0,, Y0);
        //__SET_VAR(DataIn->, DEB_ER,, Er);
        //__SET_VAR(DataIn->, DEB_CM,, Ocm);
        //__SET_VAR(DataIn->, DEB_M,, Om);
    }
}

#endif // LIB_APP_FLR



/** MBRTU MASTER
 */

#ifdef LIB_APP_MBRTU_MASTER

/** @brief  MbRtuReadBool.
 *  @param  DataIn - data.
 *  @return None.
 */
void AppMbRtuMaster_ReadBool(MBRTUREADBOOL *DataIn)
{
    if(DataIn && PlcAppFuncs.MbRtuMaster)
    {
        BOOL Ex = __GET_VAR(DataIn->EX);
        
        if(Ex)
        {
            BYTE Ok = __GET_VAR(DataIn->OK);
            if(!(Ok == PLC_APP_MBRTU_MASTER_EXEC_QUEUE || Ok == PLC_APP_MBRTU_MASTER_EXEC))
            {
                BOOL ExPrev = __GET_VAR(DataIn->EXPREV);
                plc_app_mbrtu_master_reg_t Reg;
                
                Reg.Set.DataClass = PLC_APP_MBRTU_MASTER_R_BOOL;
                Reg.Set.Com       = __GET_VAR(DataIn->COM);
                Reg.Set.Func      = __GET_VAR(DataIn->FUNC);
                Reg.Set.Num       = __GET_VAR(DataIn->NUM);
                Reg.Set.ByteOrder = 0;
                Reg.ID            = __GET_VAR(DataIn->ID);
                Reg.Addr1         = __GET_VAR(DataIn->ADDR1);
                Reg.Ok            = __GET_VAR_REF(DataIn->OK);
                Reg.Regs[0]       = (void *)__GET_VAR_REF(DataIn->O1);
                Reg.Regs[1]       = (void *)__GET_VAR_REF(DataIn->O2);
                Reg.Regs[2]       = (void *)__GET_VAR_REF(DataIn->O3);
                Reg.Regs[3]       = (void *)__GET_VAR_REF(DataIn->O4);
                Reg.Regs[4]       = (void *)__GET_VAR_REF(DataIn->O5);
                Reg.Regs[5]       = (void *)__GET_VAR_REF(DataIn->O6);
                Reg.Regs[6]       = (void *)__GET_VAR_REF(DataIn->O7);
                Reg.Regs[7]       = (void *)__GET_VAR_REF(DataIn->O8);
                
                PlcAppFuncs.MbRtuMaster(&Ex, &ExPrev, &Reg);
            }
        }
        else
        {
            __SET_VAR(DataIn->, OK,, PLC_APP_MBRTU_MASTER_OFF);
        }
        
        __SET_VAR(DataIn->, EXPREV,, Ex);
    }
}

/** @brief  MbRtuReadInt.
 *  @param  DataIn - data.
 *  @return None.
 */
void AppMbRtuMaster_ReadInt(MBRTUREADINT *DataIn)
{
    if(DataIn && PlcAppFuncs.MbRtuMaster)
    {
        BOOL Ex = __GET_VAR(DataIn->EX);
        
        if(Ex)
        {
            BYTE Ok = __GET_VAR(DataIn->OK);
            if(!(Ok == PLC_APP_MBRTU_MASTER_EXEC_QUEUE || Ok == PLC_APP_MBRTU_MASTER_EXEC))
            {
                BOOL ExPrev = __GET_VAR(DataIn->EXPREV);
                plc_app_mbrtu_master_reg_t Reg;
                
                Reg.Set.DataClass = PLC_APP_MBRTU_MASTER_R_INT;
                Reg.Set.Com       = __GET_VAR(DataIn->COM);
                Reg.Set.Func      = __GET_VAR(DataIn->FUNC);
                Reg.Set.Num       = __GET_VAR(DataIn->NUM);
                Reg.Set.ByteOrder = __GET_VAR(DataIn->BO);
                Reg.ID            = __GET_VAR(DataIn->ID);
                Reg.Addr1         = __GET_VAR(DataIn->ADDR1);
                Reg.Ok            = __GET_VAR_REF(DataIn->OK);
                Reg.Regs[0]       = (void *)__GET_VAR_REF(DataIn->O1);
                Reg.Regs[1]       = (void *)__GET_VAR_REF(DataIn->O2);
                Reg.Regs[2]       = (void *)__GET_VAR_REF(DataIn->O3);
                Reg.Regs[3]       = (void *)__GET_VAR_REF(DataIn->O4);
                Reg.Regs[4]       = (void *)__GET_VAR_REF(DataIn->O5);
                Reg.Regs[5]       = (void *)__GET_VAR_REF(DataIn->O6);
                Reg.Regs[6]       = (void *)__GET_VAR_REF(DataIn->O7);
                Reg.Regs[7]       = (void *)__GET_VAR_REF(DataIn->O8);
                
                PlcAppFuncs.MbRtuMaster(&Ex, &ExPrev, &Reg);
            }
        }
        else
        {
            __SET_VAR(DataIn->, OK,, PLC_APP_MBRTU_MASTER_OFF);
        }
        
        __SET_VAR(DataIn->, EXPREV,, Ex);
    }
}

/** @brief  MbRtuReadDInt.
 *  @param  DataIn - data.
 *  @return None.
 */
void AppMbRtuMaster_ReadDInt(MBRTUREADDINT *DataIn)
{
    if(DataIn && PlcAppFuncs.MbRtuMaster)
    {
        BOOL Ex = __GET_VAR(DataIn->EX);
        
        if(Ex)
        {
            BYTE Ok = __GET_VAR(DataIn->OK);
            if(!(Ok == PLC_APP_MBRTU_MASTER_EXEC_QUEUE || Ok == PLC_APP_MBRTU_MASTER_EXEC))
            {
                BOOL ExPrev = __GET_VAR(DataIn->EXPREV);
                plc_app_mbrtu_master_reg_t Reg;
                
                Reg.Set.DataClass = PLC_APP_MBRTU_MASTER_R_DINT;
                Reg.Set.Com       = __GET_VAR(DataIn->COM);
                Reg.Set.Func      = __GET_VAR(DataIn->FUNC);
                Reg.Set.Num       = __GET_VAR(DataIn->NUM);
                Reg.Set.ByteOrder = __GET_VAR(DataIn->BO);
                Reg.ID            = __GET_VAR(DataIn->ID);
                Reg.Addr1         = __GET_VAR(DataIn->ADDR1);
                Reg.Ok            = __GET_VAR_REF(DataIn->OK);
                Reg.Regs[0]       = (void *)__GET_VAR_REF(DataIn->O1);
                Reg.Regs[1]       = (void *)__GET_VAR_REF(DataIn->O2);
                Reg.Regs[2]       = (void *)__GET_VAR_REF(DataIn->O3);
                Reg.Regs[3]       = (void *)__GET_VAR_REF(DataIn->O4);
                Reg.Regs[4]       = (void *)__GET_VAR_REF(DataIn->O5);
                Reg.Regs[5]       = (void *)__GET_VAR_REF(DataIn->O6);
                Reg.Regs[6]       = (void *)__GET_VAR_REF(DataIn->O7);
                Reg.Regs[7]       = (void *)__GET_VAR_REF(DataIn->O8);
                
                PlcAppFuncs.MbRtuMaster(&Ex, &ExPrev, &Reg);
            }
        }
        else
        {
            __SET_VAR(DataIn->, OK,, PLC_APP_MBRTU_MASTER_OFF);
        }
        
        __SET_VAR(DataIn->, EXPREV,, Ex);
    }
}

/** @brief  MbRtuReadWord.
 *  @param  DataIn - data.
 *  @return None.
 */
void AppMbRtuMaster_ReadWord(MBRTUREADWORD *DataIn)
{
    if(DataIn && PlcAppFuncs.MbRtuMaster)
    {
        BOOL Ex = __GET_VAR(DataIn->EX);
        
        if(Ex)
        {
            BYTE Ok = __GET_VAR(DataIn->OK);
            if(!(Ok == PLC_APP_MBRTU_MASTER_EXEC_QUEUE || Ok == PLC_APP_MBRTU_MASTER_EXEC))
            {
                BOOL ExPrev = __GET_VAR(DataIn->EXPREV);
                plc_app_mbrtu_master_reg_t Reg;
                
                Reg.Set.DataClass = PLC_APP_MBRTU_MASTER_R_WORD;
                Reg.Set.Com       = __GET_VAR(DataIn->COM);
                Reg.Set.Func      = __GET_VAR(DataIn->FUNC);
                Reg.Set.Num       = __GET_VAR(DataIn->NUM);
                Reg.Set.ByteOrder = __GET_VAR(DataIn->BO);
                Reg.ID            = __GET_VAR(DataIn->ID);
                Reg.Addr1         = __GET_VAR(DataIn->ADDR1);
                Reg.Ok            = __GET_VAR_REF(DataIn->OK);
                Reg.Regs[0]       = (void *)__GET_VAR_REF(DataIn->O1);
                Reg.Regs[1]       = (void *)__GET_VAR_REF(DataIn->O2);
                Reg.Regs[2]       = (void *)__GET_VAR_REF(DataIn->O3);
                Reg.Regs[3]       = (void *)__GET_VAR_REF(DataIn->O4);
                Reg.Regs[4]       = (void *)__GET_VAR_REF(DataIn->O5);
                Reg.Regs[5]       = (void *)__GET_VAR_REF(DataIn->O6);
                Reg.Regs[6]       = (void *)__GET_VAR_REF(DataIn->O7);
                Reg.Regs[7]       = (void *)__GET_VAR_REF(DataIn->O8);
                
                PlcAppFuncs.MbRtuMaster(&Ex, &ExPrev, &Reg);
            }
        }
        else
        {
            __SET_VAR(DataIn->, OK,, PLC_APP_MBRTU_MASTER_OFF);
        }
        
        __SET_VAR(DataIn->, EXPREV,, Ex);
    }
}

/** @brief  MbRtuReadDWord.
 *  @param  DataIn - data.
 *  @return None.
 */
void AppMbRtuMaster_ReadDWord(MBRTUREADDWORD *DataIn)
{
    if(DataIn && PlcAppFuncs.MbRtuMaster)
    {
        BOOL Ex = __GET_VAR(DataIn->EX);
        
        if(Ex)
        {
            BYTE Ok = __GET_VAR(DataIn->OK);
            if(!(Ok == PLC_APP_MBRTU_MASTER_EXEC_QUEUE || Ok == PLC_APP_MBRTU_MASTER_EXEC))
            {
                BOOL ExPrev = __GET_VAR(DataIn->EXPREV);
                plc_app_mbrtu_master_reg_t Reg;
                
                Reg.Set.DataClass = PLC_APP_MBRTU_MASTER_R_DWORD;
                Reg.Set.Com       = __GET_VAR(DataIn->COM);
                Reg.Set.Func      = __GET_VAR(DataIn->FUNC);
                Reg.Set.Num       = __GET_VAR(DataIn->NUM);
                Reg.Set.ByteOrder = __GET_VAR(DataIn->BO);
                Reg.ID            = __GET_VAR(DataIn->ID);
                Reg.Addr1         = __GET_VAR(DataIn->ADDR1);
                Reg.Ok            = __GET_VAR_REF(DataIn->OK);
                Reg.Regs[0]       = (void *)__GET_VAR_REF(DataIn->O1);
                Reg.Regs[1]       = (void *)__GET_VAR_REF(DataIn->O2);
                Reg.Regs[2]       = (void *)__GET_VAR_REF(DataIn->O3);
                Reg.Regs[3]       = (void *)__GET_VAR_REF(DataIn->O4);
                Reg.Regs[4]       = (void *)__GET_VAR_REF(DataIn->O5);
                Reg.Regs[5]       = (void *)__GET_VAR_REF(DataIn->O6);
                Reg.Regs[6]       = (void *)__GET_VAR_REF(DataIn->O7);
                Reg.Regs[7]       = (void *)__GET_VAR_REF(DataIn->O8);
                
                PlcAppFuncs.MbRtuMaster(&Ex, &ExPrev, &Reg);
            }
        }
        else
        {
            __SET_VAR(DataIn->, OK,, PLC_APP_MBRTU_MASTER_OFF);
        }
        
        __SET_VAR(DataIn->, EXPREV,, Ex);
    }
}

/** @brief  MbRtuReadReal.
 *  @param  DataIn - data.
 *  @return None.
 */
void AppMbRtuMaster_ReadReal(MBRTUREADREAL *DataIn)
{
    if(DataIn && PlcAppFuncs.MbRtuMaster)
    {
        BOOL Ex = __GET_VAR(DataIn->EX);
        
        if(Ex)
        {
            BYTE Ok = __GET_VAR(DataIn->OK);
            if(!(Ok == PLC_APP_MBRTU_MASTER_EXEC_QUEUE || Ok == PLC_APP_MBRTU_MASTER_EXEC))
            {
                BOOL ExPrev = __GET_VAR(DataIn->EXPREV);
                plc_app_mbrtu_master_reg_t Reg;
                
                Reg.Set.DataClass = PLC_APP_MBRTU_MASTER_R_REAL;
                Reg.Set.Com       = __GET_VAR(DataIn->COM);
                Reg.Set.Func      = __GET_VAR(DataIn->FUNC);
                Reg.Set.Num       = __GET_VAR(DataIn->NUM);
                Reg.Set.ByteOrder = __GET_VAR(DataIn->BO);
                Reg.ID            = __GET_VAR(DataIn->ID);
                Reg.Addr1         = __GET_VAR(DataIn->ADDR1);
                Reg.Ok            = __GET_VAR_REF(DataIn->OK);
                Reg.Regs[0]       = (void *)__GET_VAR_REF(DataIn->O1);
                Reg.Regs[1]       = (void *)__GET_VAR_REF(DataIn->O2);
                Reg.Regs[2]       = (void *)__GET_VAR_REF(DataIn->O3);
                Reg.Regs[3]       = (void *)__GET_VAR_REF(DataIn->O4);
                Reg.Regs[4]       = (void *)__GET_VAR_REF(DataIn->O5);
                Reg.Regs[5]       = (void *)__GET_VAR_REF(DataIn->O6);
                Reg.Regs[6]       = (void *)__GET_VAR_REF(DataIn->O7);
                Reg.Regs[7]       = (void *)__GET_VAR_REF(DataIn->O8);
                
                PlcAppFuncs.MbRtuMaster(&Ex, &ExPrev, &Reg);
            }
        }
        else
        {
            __SET_VAR(DataIn->, OK,, PLC_APP_MBRTU_MASTER_OFF);
        }
        
        __SET_VAR(DataIn->, EXPREV,, Ex);
    }
}

/** @brief  MbRtuWriteBool.
 *  @param  DataIn - data.
 *  @return None.
 */
void AppMbRtuMaster_WriteBool(MBRTUWRITEBOOL *DataIn)
{
    if(DataIn && PlcAppFuncs.MbRtuMaster)
    {
        BOOL Ex = __GET_VAR(DataIn->EX);
        
        if(Ex)
        {
            BYTE Ok = __GET_VAR(DataIn->OK);
            if(!(Ok == PLC_APP_MBRTU_MASTER_EXEC_QUEUE || Ok == PLC_APP_MBRTU_MASTER_EXEC))
            {
                BOOL ExPrev = __GET_VAR(DataIn->EXPREV);
                plc_app_mbrtu_master_reg_t Reg;
                
                Reg.Set.DataClass = PLC_APP_MBRTU_MASTER_W_BOOL;
                Reg.Set.Com       = __GET_VAR(DataIn->COM);
                Reg.Set.Func      = __GET_VAR(DataIn->FUNC);
                Reg.Set.Num       = __GET_VAR(DataIn->NUM);
                Reg.Set.ByteOrder = 0;
                Reg.ID            = __GET_VAR(DataIn->ID);
                Reg.Addr1         = __GET_VAR(DataIn->ADDR1);
                Reg.Ok            = __GET_VAR_REF(DataIn->OK);
                Reg.Regs[0]       = (void *)__GET_VAR_REF(DataIn->I1);
                Reg.Regs[1]       = (void *)__GET_VAR_REF(DataIn->I2);
                Reg.Regs[2]       = (void *)__GET_VAR_REF(DataIn->I3);
                Reg.Regs[3]       = (void *)__GET_VAR_REF(DataIn->I4);
                Reg.Regs[4]       = (void *)__GET_VAR_REF(DataIn->I5);
                Reg.Regs[5]       = (void *)__GET_VAR_REF(DataIn->I6);
                Reg.Regs[6]       = (void *)__GET_VAR_REF(DataIn->I7);
                Reg.Regs[7]       = (void *)__GET_VAR_REF(DataIn->I8);
                
                PlcAppFuncs.MbRtuMaster(&Ex, &ExPrev, &Reg);
            }
        }
        else
        {
            __SET_VAR(DataIn->, OK,, PLC_APP_MBRTU_MASTER_OFF);
        }
        
        __SET_VAR(DataIn->, EXPREV,, Ex);
    }
}

/** @brief  MbRtuWriteInt.
 *  @param  DataIn - data.
 *  @return None.
 */
void AppMbRtuMaster_WriteInt(MBRTUWRITEINT *DataIn)
{
    if(DataIn && PlcAppFuncs.MbRtuMaster)
    {
        BOOL Ex = __GET_VAR(DataIn->EX);
        
        if(Ex)
        {
            BYTE Ok = __GET_VAR(DataIn->OK);
            if(!(Ok == PLC_APP_MBRTU_MASTER_EXEC_QUEUE || Ok == PLC_APP_MBRTU_MASTER_EXEC))
            {
                BOOL ExPrev = __GET_VAR(DataIn->EXPREV);
                plc_app_mbrtu_master_reg_t Reg;
                
                Reg.Set.DataClass = PLC_APP_MBRTU_MASTER_W_INT;
                Reg.Set.Com       = __GET_VAR(DataIn->COM);
                Reg.Set.Func      = __GET_VAR(DataIn->FUNC);
                Reg.Set.Num       = __GET_VAR(DataIn->NUM);
                Reg.Set.ByteOrder = __GET_VAR(DataIn->BO);
                Reg.ID            = __GET_VAR(DataIn->ID);
                Reg.Addr1         = __GET_VAR(DataIn->ADDR1);
                Reg.Ok            = __GET_VAR_REF(DataIn->OK);
                Reg.Regs[0]       = (void *)__GET_VAR_REF(DataIn->I1);
                Reg.Regs[1]       = (void *)__GET_VAR_REF(DataIn->I2);
                Reg.Regs[2]       = (void *)__GET_VAR_REF(DataIn->I3);
                Reg.Regs[3]       = (void *)__GET_VAR_REF(DataIn->I4);
                Reg.Regs[4]       = (void *)__GET_VAR_REF(DataIn->I5);
                Reg.Regs[5]       = (void *)__GET_VAR_REF(DataIn->I6);
                Reg.Regs[6]       = (void *)__GET_VAR_REF(DataIn->I7);
                Reg.Regs[7]       = (void *)__GET_VAR_REF(DataIn->I8);
                
                PlcAppFuncs.MbRtuMaster(&Ex, &ExPrev, &Reg);
            }
        }
        else
        {
            __SET_VAR(DataIn->, OK,, PLC_APP_MBRTU_MASTER_OFF);
        }
        
        __SET_VAR(DataIn->, EXPREV,, Ex);
    }
}

/** @brief  MbRtuWriteDInt.
 *  @param  DataIn - data.
 *  @return None.
 */
void AppMbRtuMaster_WriteDInt(MBRTUWRITEDINT *DataIn)
{
    if(DataIn && PlcAppFuncs.MbRtuMaster)
    {
        BOOL Ex = __GET_VAR(DataIn->EX);
        
        if(Ex)
        {
            BYTE Ok = __GET_VAR(DataIn->OK);
            if(!(Ok == PLC_APP_MBRTU_MASTER_EXEC_QUEUE || Ok == PLC_APP_MBRTU_MASTER_EXEC))
            {
                BOOL ExPrev = __GET_VAR(DataIn->EXPREV);
                plc_app_mbrtu_master_reg_t Reg;
                
                Reg.Set.DataClass = PLC_APP_MBRTU_MASTER_W_DINT;
                Reg.Set.Com       = __GET_VAR(DataIn->COM);
                Reg.Set.Func      = __GET_VAR(DataIn->FUNC);
                Reg.Set.Num       = __GET_VAR(DataIn->NUM);
                Reg.Set.ByteOrder = __GET_VAR(DataIn->BO);
                Reg.ID            = __GET_VAR(DataIn->ID);
                Reg.Addr1         = __GET_VAR(DataIn->ADDR1);
                Reg.Ok            = __GET_VAR_REF(DataIn->OK);
                Reg.Regs[0]       = (void *)__GET_VAR_REF(DataIn->I1);
                Reg.Regs[1]       = (void *)__GET_VAR_REF(DataIn->I2);
                Reg.Regs[2]       = (void *)__GET_VAR_REF(DataIn->I3);
                Reg.Regs[3]       = (void *)__GET_VAR_REF(DataIn->I4);
                Reg.Regs[4]       = (void *)__GET_VAR_REF(DataIn->I5);
                Reg.Regs[5]       = (void *)__GET_VAR_REF(DataIn->I6);
                Reg.Regs[6]       = (void *)__GET_VAR_REF(DataIn->I7);
                Reg.Regs[7]       = (void *)__GET_VAR_REF(DataIn->I8);
                
                PlcAppFuncs.MbRtuMaster(&Ex, &ExPrev, &Reg);
            }
        }
        else
        {
            __SET_VAR(DataIn->, OK,, PLC_APP_MBRTU_MASTER_OFF);
        }
        
        __SET_VAR(DataIn->, EXPREV,, Ex);
    }
}

/** @brief  MbRtuWriteWord.
 *  @param  DataIn - data.
 *  @return None.
 */
void AppMbRtuMaster_WriteWord(MBRTUWRITEWORD *DataIn)
{
    if(DataIn && PlcAppFuncs.MbRtuMaster)
    {
        BOOL Ex = __GET_VAR(DataIn->EX);
        
        if(Ex)
        {
            BYTE Ok = __GET_VAR(DataIn->OK);
            if(!(Ok == PLC_APP_MBRTU_MASTER_EXEC_QUEUE || Ok == PLC_APP_MBRTU_MASTER_EXEC))
            {
                BOOL ExPrev = __GET_VAR(DataIn->EXPREV);
                plc_app_mbrtu_master_reg_t Reg;
                
                Reg.Set.DataClass = PLC_APP_MBRTU_MASTER_W_WORD;
                Reg.Set.Com       = __GET_VAR(DataIn->COM);
                Reg.Set.Func      = __GET_VAR(DataIn->FUNC);
                Reg.Set.Num       = __GET_VAR(DataIn->NUM);
                Reg.Set.ByteOrder = __GET_VAR(DataIn->BO);
                Reg.ID            = __GET_VAR(DataIn->ID);
                Reg.Addr1         = __GET_VAR(DataIn->ADDR1);
                Reg.Ok            = __GET_VAR_REF(DataIn->OK);
                Reg.Regs[0]       = (void *)__GET_VAR_REF(DataIn->I1);
                Reg.Regs[1]       = (void *)__GET_VAR_REF(DataIn->I2);
                Reg.Regs[2]       = (void *)__GET_VAR_REF(DataIn->I3);
                Reg.Regs[3]       = (void *)__GET_VAR_REF(DataIn->I4);
                Reg.Regs[4]       = (void *)__GET_VAR_REF(DataIn->I5);
                Reg.Regs[5]       = (void *)__GET_VAR_REF(DataIn->I6);
                Reg.Regs[6]       = (void *)__GET_VAR_REF(DataIn->I7);
                Reg.Regs[7]       = (void *)__GET_VAR_REF(DataIn->I8);
                
                PlcAppFuncs.MbRtuMaster(&Ex, &ExPrev, &Reg);
            }
        }
        else
        {
            __SET_VAR(DataIn->, OK,, PLC_APP_MBRTU_MASTER_OFF);
        }
        
        __SET_VAR(DataIn->, EXPREV,, Ex);
    }
}

/** @brief  MbRtuWriteDWord.
 *  @param  DataIn - data.
 *  @return None.
 */
void AppMbRtuMaster_WriteDWord(MBRTUWRITEDWORD *DataIn)
{
    if(DataIn && PlcAppFuncs.MbRtuMaster)
    {
        BOOL Ex = __GET_VAR(DataIn->EX);
        
        if(Ex)
        {
            BYTE Ok = __GET_VAR(DataIn->OK);
            if(!(Ok == PLC_APP_MBRTU_MASTER_EXEC_QUEUE || Ok == PLC_APP_MBRTU_MASTER_EXEC))
            {
                BOOL ExPrev = __GET_VAR(DataIn->EXPREV);
                plc_app_mbrtu_master_reg_t Reg;
                
                Reg.Set.DataClass = PLC_APP_MBRTU_MASTER_W_DWORD;
                Reg.Set.Com       = __GET_VAR(DataIn->COM);
                Reg.Set.Func      = __GET_VAR(DataIn->FUNC);
                Reg.Set.Num       = __GET_VAR(DataIn->NUM);
                Reg.Set.ByteOrder = __GET_VAR(DataIn->BO);
                Reg.ID            = __GET_VAR(DataIn->ID);
                Reg.Addr1         = __GET_VAR(DataIn->ADDR1);
                Reg.Ok            = __GET_VAR_REF(DataIn->OK);
                Reg.Regs[0]       = (void *)__GET_VAR_REF(DataIn->I1);
                Reg.Regs[1]       = (void *)__GET_VAR_REF(DataIn->I2);
                Reg.Regs[2]       = (void *)__GET_VAR_REF(DataIn->I3);
                Reg.Regs[3]       = (void *)__GET_VAR_REF(DataIn->I4);
                Reg.Regs[4]       = (void *)__GET_VAR_REF(DataIn->I5);
                Reg.Regs[5]       = (void *)__GET_VAR_REF(DataIn->I6);
                Reg.Regs[6]       = (void *)__GET_VAR_REF(DataIn->I7);
                Reg.Regs[7]       = (void *)__GET_VAR_REF(DataIn->I8);
                
                PlcAppFuncs.MbRtuMaster(&Ex, &ExPrev, &Reg);
            }
        }
        else
        {
            __SET_VAR(DataIn->, OK,, PLC_APP_MBRTU_MASTER_OFF);
        }
        
        __SET_VAR(DataIn->, EXPREV,, Ex);
    }
}

/** @brief  MbRtuWriteReal.
 *  @param  DataIn - data.
 *  @return None.
 */
void AppMbRtuMaster_WriteReal(MBRTUWRITEREAL *DataIn)
{
    if(DataIn && PlcAppFuncs.MbRtuMaster)
    {
        BOOL Ex = __GET_VAR(DataIn->EX);
        
        if(Ex)
        {
            BYTE Ok = __GET_VAR(DataIn->OK);
            if(!(Ok == PLC_APP_MBRTU_MASTER_EXEC_QUEUE || Ok == PLC_APP_MBRTU_MASTER_EXEC))
            {
                BOOL ExPrev = __GET_VAR(DataIn->EXPREV);
                plc_app_mbrtu_master_reg_t Reg;
                
                Reg.Set.DataClass = PLC_APP_MBRTU_MASTER_W_REAL;
                Reg.Set.Com       = __GET_VAR(DataIn->COM);
                Reg.Set.Func      = __GET_VAR(DataIn->FUNC);
                Reg.Set.Num       = __GET_VAR(DataIn->NUM);
                Reg.Set.ByteOrder = __GET_VAR(DataIn->BO);
                Reg.ID            = __GET_VAR(DataIn->ID);
                Reg.Addr1         = __GET_VAR(DataIn->ADDR1);
                Reg.Ok            = __GET_VAR_REF(DataIn->OK);
                Reg.Regs[0]       = (void *)__GET_VAR_REF(DataIn->I1);
                Reg.Regs[1]       = (void *)__GET_VAR_REF(DataIn->I2);
                Reg.Regs[2]       = (void *)__GET_VAR_REF(DataIn->I3);
                Reg.Regs[3]       = (void *)__GET_VAR_REF(DataIn->I4);
                Reg.Regs[4]       = (void *)__GET_VAR_REF(DataIn->I5);
                Reg.Regs[5]       = (void *)__GET_VAR_REF(DataIn->I6);
                Reg.Regs[6]       = (void *)__GET_VAR_REF(DataIn->I7);
                Reg.Regs[7]       = (void *)__GET_VAR_REF(DataIn->I8);
                
                PlcAppFuncs.MbRtuMaster(&Ex, &ExPrev, &Reg);
            }
        }
        else
        {
            __SET_VAR(DataIn->, OK,, PLC_APP_MBRTU_MASTER_OFF);
        }
        
        __SET_VAR(DataIn->, EXPREV,, Ex);
    }
}

#endif // LIB_APP_MBRTU_MASTER



/** SYSTEM
 */

#ifdef LIB_APP_SYS

/** @brief  Software reset.
 *  @param Ex - allow to execution:
 *  @arg     = false - not allow
 *  @arg     = true  - allow
 *  @return None.
 */
void AppSys_SoftReset(BOOL ExIn)
{
    if(PlcAppFuncs.SoftReset)
    {
        PlcAppFuncs.SoftReset(ExIn);
    }
}

#endif // LIB_APP_SYS
/**
 * Tail of code common to all C targets
 **/

/** 
 * LOGGING
 **/

#ifndef LOG_BUFFER_SIZE
#define LOG_BUFFER_SIZE (1<<14) /*16Ko*/
#endif
#ifndef LOG_BUFFER_ATTRS
#define LOG_BUFFER_ATTRS
#endif

#define LOG_BUFFER_MASK (LOG_BUFFER_SIZE-1)

static char LogBuff[LOG_LEVELS][LOG_BUFFER_SIZE] LOG_BUFFER_ATTRS;
static void inline copy_to_log(uint8_t level, uint32_t buffpos, void* buf, uint32_t size){
    if(buffpos + size < LOG_BUFFER_SIZE){
        memcpy(&LogBuff[level][buffpos], buf, size);
    }else{
        uint32_t remaining = LOG_BUFFER_SIZE - buffpos; 
        memcpy(&LogBuff[level][buffpos], buf, remaining);
        memcpy(LogBuff[level], (char*)buf + remaining, size - remaining);
    }
}
static void inline copy_from_log(uint8_t level, uint32_t buffpos, void* buf, uint32_t size){
    if(buffpos + size < LOG_BUFFER_SIZE){
        memcpy(buf, &LogBuff[level][buffpos], size);
    }else{
        uint32_t remaining = LOG_BUFFER_SIZE - buffpos; 
        memcpy(buf, &LogBuff[level][buffpos], remaining);
        memcpy((char*)buf + remaining, LogBuff[level], size - remaining);
    }
}

/* Log buffer structure

 |<-Tail1.msgsize->|<-sizeof(mTail)->|<--Tail2.msgsize-->|<-sizeof(mTail)->|...
 |  Message1 Body  |      Tail1      |   Message2 Body   |      Tail2      |

*/
typedef struct {
    uint32_t msgidx;
    uint32_t msgsize;
    unsigned long tick;
    IEC_TIME time;
} mTail;

/* Log cursor : 64b
   |63 ... 32|31 ... 0|
   | Message | Buffer |
   | counter | Index  | */
static uint64_t LogCursor[LOG_LEVELS] LOG_BUFFER_ATTRS = {0x0,0x0,0x0,0x0};

void ResetLogCount(void) {
	uint8_t level;
	for(level=0;level<LOG_LEVELS;level++){
		LogCursor[level] = 0;
	}
}

/* Store one log message of give size */
int LogMessage(uint8_t level, char* buf, uint32_t size){
    if(size < LOG_BUFFER_SIZE - sizeof(mTail)){
        uint32_t buffpos;
        uint64_t new_cursor, old_cursor;

        mTail tail;
        tail.msgsize = size;
        tail.tick = __tick;
        PLC_GetTime(&tail.time);

        /* We cannot increment both msg index and string pointer 
           in a single atomic operation but we can detect having been interrupted.
           So we can try with atomic compare and swap in a loop until operation
           succeeds non interrupted */
        do{
            old_cursor = LogCursor[level];
            buffpos = (uint32_t)old_cursor;
            tail.msgidx = (old_cursor >> 32); 
            new_cursor = ((uint64_t)(tail.msgidx + 1)<<32) 
                         | (uint64_t)((buffpos + size + sizeof(mTail)) & LOG_BUFFER_MASK);
        }while(AtomicCompareExchange64(
            (long long*)&LogCursor[level],
            (long long)old_cursor,
            (long long)new_cursor)!=(long long)old_cursor);

        copy_to_log(level, buffpos, buf, size);
        copy_to_log(level, (buffpos + size) & LOG_BUFFER_MASK, &tail, sizeof(mTail));

        return 1; /* Success */
    }else{
    	char mstr[] = "Logging error : message too big";
        LogMessage(LOG_CRITICAL, mstr, sizeof(mstr));
    }
    return 0;
}

uint32_t GetLogCount(uint8_t level){
    return (uint64_t)LogCursor[level] >> 32;
}

/* Return message size and content */
uint32_t GetLogMessage(uint8_t level, uint32_t msgidx, char* buf, uint32_t max_size, uint32_t* tick, uint32_t* tv_sec, uint32_t* tv_nsec){
    uint64_t cursor = LogCursor[level];
    if(cursor){
        /* seach cursor */
        uint32_t stailpos = (uint32_t)cursor; 
        uint32_t smsgidx;
        mTail tail;
        tail.msgidx = cursor >> 32;
        tail.msgsize = 0;

        /* Message search loop */
        do {
            smsgidx = tail.msgidx;
            stailpos = (stailpos - sizeof(mTail) - tail.msgsize ) & LOG_BUFFER_MASK;
            copy_from_log(level, stailpos, &tail, sizeof(mTail));
        }while((tail.msgidx == smsgidx - 1) && (tail.msgidx > msgidx));

        if(tail.msgidx == msgidx){
            uint32_t sbuffpos = (stailpos - tail.msgsize ) & LOG_BUFFER_MASK; 
            uint32_t totalsize = tail.msgsize;
            *tick = tail.tick; 
            *tv_sec = tail.time.tv_sec; 
            *tv_nsec = tail.time.tv_nsec; 
            copy_from_log(level, sbuffpos, buf, 
                          totalsize > max_size ? max_size : totalsize);
            return totalsize;
        }
    }
    return 0;
}

#define CALIBRATED -2
#define NOT_CALIBRATED -1
static int calibration_count = NOT_CALIBRATED;
static IEC_TIME cal_begin;
static long long Tsync = 0;
static long long FreqCorr = 0;
static int Nticks = 0;
static unsigned long last_tick = 0;

/*
 * Called on each external periodic sync event
 * make PLC tick synchronous with external sync
 * ratio defines when PLC tick occurs between two external sync
 * @param sync_align_ratio 
 *          0->100 : align ratio
 *          < 0 : no align, calibrate period
 **/
void align_tick(int sync_align_ratio)
{
	/*
	printf("align_tick(%d)\n", calibrate);
	*/
	if(sync_align_ratio < 0){ /* Calibration */
		if(calibration_count == CALIBRATED)
			/* Re-calibration*/
			calibration_count = NOT_CALIBRATED;
		if(calibration_count == NOT_CALIBRATED)
			/* Calibration start, get time*/
			PLC_GetTime(&cal_begin);
		calibration_count++;
	}else{ /* do alignment (if possible) */
		if(calibration_count >= 0){
			/* End of calibration */
			/* Get final time */
			IEC_TIME cal_end;
			PLC_GetTime(&cal_end);
			/*adjust calibration_count*/
			calibration_count++;
			/* compute mean of Tsync, over calibration period */
			Tsync = ((long long)(cal_end.tv_sec - cal_begin.tv_sec) * (long long)1000000000 +
					(cal_end.tv_nsec - cal_begin.tv_nsec)) / calibration_count;
			if( (Nticks = (Tsync / common_ticktime__)) > 0){
				FreqCorr = (Tsync % common_ticktime__); /* to be divided by Nticks */
			}else{
				FreqCorr = Tsync - (common_ticktime__ % Tsync);
			}
			/*
			printf("Tsync = %ld\n", Tsync);
			printf("calibration_count = %d\n", calibration_count);
			printf("Nticks = %d\n", Nticks);
			*/
			calibration_count = CALIBRATED;
		}
		if(calibration_count == CALIBRATED){
			/* Get Elapsed time since last PLC tick (__CURRENT_TIME) */
			IEC_TIME now;
			long long elapsed;
			long long Tcorr;
			long long PhaseCorr;
			long long PeriodicTcorr;
			PLC_GetTime(&now);
			elapsed = (now.tv_sec - __CURRENT_TIME.tv_sec) * 1000000000 + now.tv_nsec - __CURRENT_TIME.tv_nsec;
			if(Nticks > 0){
				PhaseCorr = elapsed - (common_ticktime__ + FreqCorr/Nticks)*sync_align_ratio/100; /* to be divided by Nticks */
				Tcorr = common_ticktime__ + (PhaseCorr + FreqCorr) / Nticks;
				if(Nticks < 2){
					/* When Sync source period is near Tick time */
					/* PhaseCorr may not be applied to Periodic time given to timer */
					PeriodicTcorr = common_ticktime__ + FreqCorr / Nticks;
				}else{
					PeriodicTcorr = Tcorr;
				}
			}else if(__tick > last_tick){
				last_tick = __tick;
				PhaseCorr = elapsed - (Tsync*sync_align_ratio/100);
				PeriodicTcorr = Tcorr = common_ticktime__ + PhaseCorr + FreqCorr;
			}else{
				/*PLC did not run meanwhile. Nothing to do*/
				return;
			}
			/* DO ALIGNEMENT */
			PLC_SetTimer(Tcorr - elapsed, PeriodicTcorr);
		}
	}
}
